@@ -1,3 +1,5 @@
+__VERSION__ = "2.13.1"
+
 import collections
 import functools
 import hashlib
@@ -30,7 +32,7 @@ _user_array_setters = {
 
 def stringify(content):
     if isinstance(content, str):
-        if content[0] in "'\"" and content[-1] == content[0] and len(content) > 1:
+        if len(content) > 1 and content[0] in "'\"" and content[-1] == content[0]:
             return content
 
         return '"' + repr(content)[1:-1] + '"'
@@ -469,28 +471,121 @@ class ZDClassTemplate(ZDBaseActor):
         self.abstract_array_names = dict(abstract_array_names)
 
         self.parse_data = parse_data
-        self.parametric_table = {}
+        self.existing = {}
 
-    def generated_class_name(self, parameter_values, new_id):
-        hash = hashlib.sha1()
+    def duplicate(
+        self,
+        parameter_values,
+        provided_label_names,
+        provided_macro_names,
+        provided_array_names,
+    ):
+        if (
+            self.abstract_macro_names
+            or self.abstract_label_names
+            or self.abstract_array_names
+        ):
+            return False
+
+        return (
+            self.parameter_hash(
+                parameter_values,
+                provided_label_names,
+                provided_macro_names,
+                provided_array_names,
+            )
+            in self.existing
+        )
+
+    def which_duplicate(
+        self,
+        parameter_values,
+        provided_label_names,
+        provided_macro_names,
+        provided_array_names,
+    ):
+        return self.existing[
+            self.parameter_hash(
+                parameter_values,
+                provided_label_names,
+                provided_macro_names,
+                provided_array_names,
+            )
+        ]
+
+    def register(
+        self,
+        parameter_values,
+        provided_label_names,
+        provided_macro_names,
+        provided_array_names,
+        result,
+    ):
+        self.existing[
+            self.parameter_hash(
+                parameter_values,
+                provided_label_names,
+                provided_macro_names,
+                provided_array_names,
+            )
+        ] = result
+
+    def parameter_hash(
+        self,
+        parameter_values,
+        provided_label_names,
+        provided_macro_names,
+        provided_array_names,
+    ) -> str:
+        hash = hashlib.sha256()
 
         hash.update(self.name.encode("utf-8"))
+        hash.update(b"|")
+
         hash.update(self.id.encode("utf-8"))
-        hash.update(new_id.encode("utf-8"))
+        hash.update(b"|")
 
-        for parm in parameter_values:
-            hash.update(parm.encode("utf-8"))
+        if (
+            self.abstract_macro_names
+            or self.abstract_label_names
+            or self.abstract_array_names
+        ):
+            hash.update(make_id(80).encode("utf-8"))
 
-        for name in self.abstract_label_names:
-            hash.update(name.encode("utf-8"))
+        else:
+            for parm in parameter_values:
+                hash.update(parm.encode("utf-8"))
+                hash.update(b"-")
 
-        for name, args in self.abstract_macro_names.items():
-            hash.update(name.encode("utf-8"))
+            hash.update(b"|")
 
-            for arg in args:
-                hash.update(arg.encode("utf-8"))
+            for name in itertools.chain(provided_label_names, provided_array_names):
+                hash.update(name.encode("utf-8"))
+                hash.update(b"-")
 
-        return "{}__deriv_{}".format(self.name, hash.hexdigest())
+            hash.update(b"|")
+
+            for name, args in provided_macro_names.items():
+                hash.update(hex(len(args)).encode("utf-8"))
+                hash.update(name.encode("utf-8"))
+                hash.update(b"-")
+
+        return hash.hexdigest()
+
+    def generated_class_name(
+        self,
+        parameter_values,
+        provided_label_names,
+        provided_macro_names,
+        provided_array_names,
+    ):
+        hash = self.parameter_hash(
+            parameter_values,
+            provided_label_names,
+            provided_macro_names,
+            provided_array_names,
+        )
+        return "{}__deriv_{}".format(self.name, hash)
 
     def generate_init_class(
         self,
@@ -503,21 +598,31 @@ class ZDClassTemplate(ZDBaseActor):
         name=None,
         pending=None,
     ):
-        if (
-            tuple(parameter_values) in self.parametric_table
-            and not self.abstract_label_names
-            and not self.abstract_macro_names
-            and not self.abstract_array_names
-        ):
-            return (False, self.parametric_table[tuple(parameter_values)])
-
         provided_label_names = set(provided_label_names)
         provided_macro_names = dict(provided_macro_names)
 
+        if self.duplicate(
+            parameter_values,
+            provided_label_names,
+            provided_macro_names,
+            provided_array_names,
+        ):
+            return False, self.which_duplicate(
+                parameter_values,
+                provided_label_names,
+                provided_macro_names,
+                provided_array_names,
+            )
+
         new_name = (
             name
             if name is not None
-            else self.generated_class_name(parameter_values, make_id(40))
+            else self.generated_class_name(
+                parameter_values,
+                provided_label_names,
+                provided_macro_names,
+                provided_array_names,
+            )
         )
 
         if self.group_name:
@@ -588,9 +693,18 @@ class ZDClassTemplate(ZDBaseActor):
                     )
                 )
 
-        self.parametric_table[tuple(parameter_values)] = res
+        self.register(
+            parameter_values,
+            provided_label_names,
+            provided_macro_names,
+            provided_array_names,
+            res,
+        )
 
-        return (True, res)
+        self.code.actor_names[res.name.upper()] = res
+        context.add_actor(res)
+
+        return True, res
 
     def get_init_replacements(self, parameter_values):
         return dict(
@@ -997,28 +1111,43 @@ class CompilerError(Exception):
 class ZDCodeParseContext(object):
     def __init__(
         self,
-        replacements=(),
-        macros=(),
-        templates=(),
-        actors=(),
-        mods=(),
+        replacements=None,
+        macros=None,
+        templates=None,
+        actors=None,
+        mods=None,
         applied_mods=None,
         remote_offset=0,
+        description=None,
     ):
-        self.macros = collections.ChainMap({}, macros)
-        self.replacements = collections.ChainMap({}, replacements)
-        self.templates = collections.ChainMap({}, templates)
-        self.mods = collections.ChainMap({}, mods)
+        self.replacements = (
+            replacements.new_child()
+            if replacements is not None
+            else collections.ChainMap({})
+        )
+        self.macros = (
+            macros.new_child() if macros is not None else collections.ChainMap({})
+        )
+        self.templates = (
+            templates.new_child() if templates is not None else collections.ChainMap({})
+        )
+        self.mods = mods.new_child() if mods is not None else collections.ChainMap({})
 
         self.always_applied_mods = applied_mods
         self.applied_mods = []
 
-        self.actor_lists = list(actors) if actors else [[]]
+        self.actor_lists = list(actors) if actors else []
         self.desc_stack = []
         self.states = []
         self.remote_children = []
         self.remote_offset = remote_offset
 
+        if description:
+            self.add_description(description)
+
+    def add_description(self, desc):
+        self.desc_stack.append(desc)
+
     def get_applied_mods(self):
         if self.always_applied_mods is None:
             return iter(self.applied_mods)
@@ -1156,10 +1285,10 @@ class ZDCodeParseContext(object):
         return ZDCtxDescBlock(self, desc)
 
     def update(self, other_ctx: "ZDCodeParseContext"):
-        self.macros.update(other_ctx.macros)
-        self.replacements.update(other_ctx.replacements)
-        self.templates.update(other_ctx.templates)
-        self.mods.update(other_ctx.mods)
+        self.macros.maps.insert(-1, other_ctx.macros)
+        self.replacements.maps.insert(-1, other_ctx.replacements)
+        self.templates.maps.insert(-1, other_ctx.templates)
+        self.mods.maps.insert(-1, other_ctx.mods)
 
     def add_actor(self, ac: ZDActor):
         for al in self.actor_lists:
@@ -1311,7 +1440,10 @@ class ZDCode:
             except CompilerError as err:
                 if error_handler:
                     error_handler(err)
-                return False
+                    return False
+
+                else:
+                    raise
 
         return bool(data)
 
@@ -1326,16 +1458,24 @@ class ZDCode:
         etype, econt = evaluation
 
         if etype == "literal":
-            return econt
+            return self._parse_literal_numeric(econt, context)
 
         if etype == "operation":
             return self._parse_operation(econt, context)
 
+        raise CompilerError(
+            "Could not parse evaluation {} in {}".format(
+                repr(evaluation), context.describe()
+            )
+        )
+
     def _parse_expression(self, expr, context):
         etype, exval = expr
 
         if etype == "expr":
-            return " ".join(self._parse_expression(item, context) for item in exval)
+            return " ".join(
+                str(self._parse_expression(item, context)) for item in exval
+            )
 
         if etype == "literal":
             return self._parse_literal(exval, context)
@@ -1343,12 +1483,16 @@ class ZDCode:
         if etype == "array index":
             return "[" + self._parse_expression(exval, context) + "]"
 
-        if etype == "oper":
+        if etype == "operator":
             return exval
 
         if etype == "paren expr":
             return "(" + self._parse_expression(exval, context) + ")"
 
+        raise CompilerError(
+            "Could not parse expression {} in {}".format(repr(expr), context.describe())
+        )
+
     def _parse_argument(self, arg, context, name=None):
         atype, aval = arg
 
@@ -1380,6 +1524,32 @@ class ZDCode:
 
         return stringify(name)
 
+    def _parse_literal_numeric(self, literal, context):
+        if literal[0] == "number":
+            return literal[1]
+
+        if literal[0] == "eval":
+            return self._parse_evaluation(literal[1], context)
+
+        if literal[0] == "format string":
+            return float(self._parse_formatted_string(literal[1], context))
+
+        if literal[0] == "actor variable":
+            if literal[1].upper() in context.replacements:
+                return float(context.replacements[literal[1].upper()])
+
+            raise CompilerError(
+                "Cannot get compile-time variable for evaluation, {}, in {}".format(
+                    repr(literal[1]), context.describe()
+                )
+            )
+
+        raise CompilerError(
+            "Could not parse numeric expression {} at {}".format(
+                repr(literal), context.describe()
+            )
+        )
+
     def _parse_literal(self, literal, context):
         if isinstance(literal, str):
             return literal
@@ -1390,6 +1560,9 @@ class ZDCode:
         if literal[0] == "string":
             return stringify(literal[1])
 
+        if literal[0] == "eval":
+            return self._parse_evaluation(literal[1], context)
+
         if literal[0] == "format string":
             return stringify(self._parse_formatted_string(literal[1], context))
 
@@ -1409,6 +1582,10 @@ class ZDCode:
         if literal[0] == "template derivation":
             return self._parse_template_derivation(literal[1], context)
 
+        raise CompilerError(
+            "Could not parse literal {} in {}".format(repr(literal), context.describe())
+        )
+
     def _parse_array(self, arr, context):
         arr = dict(arr)
         arr["value"] = (
@@ -1566,6 +1743,19 @@ class ZDCode:
 
         return "".join(res)
 
+    def _parse_formattable_string(self, cval, context: ZDCodeParseContext):
+        if cval[0] == "string":
+            return cval[1]
+
+        elif cval[0] == "format":
+            return self._parse_formatted_string(cval[1], context)
+
+        raise CompilerError(
+            "Could not parse formattable string {} in {}".format(
+                repr(cval), context.describe()
+            )
+        )
+
     def _parse_replaceable_number(self, cval, context: ZDCodeParseContext):
         if isinstance(cval, str):
             cval = context.replacements.get(cval.upper(), cval)
@@ -2151,13 +2341,16 @@ class ZDCode:
         if ptype == "classname":
             return context.replacements.get(pval.upper(), pval)
 
+        elif ptype == "format":
+            return self._parse_formatted_string(pval, context)
+
         elif ptype == "template derivation":
             with context.desc_block("template derivation inheritance"):
                 return self._parse_template_derivation(
                     pval, context, do_stringify=False
                 ).name
 
-    def _parse_anonym_class(self, anonym_class, context):
+    def _parse_anonym_class(self, anonym_class, context) -> str:
         a = dict(anonym_class)
         new_context = context.derive("anonymous class")
 
@@ -2188,7 +2381,8 @@ class ZDCode:
         context.add_actor(anonym_actor)
 
         self.anonymous_classes.append(anonym_actor)
-        self.inventories.append(anonym_actor)
+        # self.actors.append(anonym_actor)
+        # self.inventories.append(anonym_actor)
 
         return stringify(anonym_actor.name)
 
@@ -2209,15 +2403,13 @@ class ZDCode:
         arrays = dict(arrays)
         body = list(body)
 
-        name = name or template.generated_class_name(param_values, make_id(40))
-
         needs_init, actor = template.generate_init_class(
             self,
             context,
             param_values,
             {l.upper() for l in labels.keys()},
             {m["name"].upper(): m["args"] for m in macros.values()},
-            {a["name"].upper(): len(a["value"]) for a in arrays.values()},
+            {a["name"].upper(): len(a["value"][1]) for a in arrays.values()},
             name=name,
         )
         new_context = actor.get_context()
@@ -2266,13 +2458,28 @@ class ZDCode:
             else:
                 pending_oper_gen().func()
 
-            self.actors.append(actor)
-            self.actor_names[actor.name.upper()] = actor
-
         return actor
 
+    def _parse_var_value(self, vval, context):
+        vvtype, vvbody = vval
+
+        if vvtype == "val":
+            return ("val", self._parse_expression(vvbody, context))
+
+        if vvtype == "arr":
+            return self._parse_array(vval, context)
+
+        raise CompilerError(
+            "Could not parse user var value {} in {}".format(
+                repr(vval), context.describe()
+            )
+        )
+
     def _parse_class_body(self, actor, context, body):
-        assert context is actor.context
+        for btype, bdata in body:
+            if btype == "for":
+                for ctx, unpacked_bdata in self.resolve_for(bdata, context):
+                    self._parse_class_body(actor, ctx, [unpacked_bdata])
 
         for btype, bdata in body:
             if btype == "mod":
@@ -2301,7 +2508,7 @@ class ZDCode:
             elif btype == "user var":
                 bdata = {
                     **bdata,
-                    "value": self._parse_expression(bdata["value"], context),
+                    "value": self._parse_var_value(bdata["value"], context),
                 }
                 actor.uservars.append(bdata)
 
@@ -2346,65 +2553,67 @@ class ZDCode:
 
                 context.applied_mods.extend(mod)
 
+    # unpack static for loops
+    def resolve_for(self, forbody, context):
+        itername, iteridx, itermode, f_body, f_else = forbody
+
+        def do_for(iterator):
+            break_ctx = context.remote_derive("static for")
+
+            for i, item in enumerate(iterator):
+                iter_ctx = break_ctx.remote_derive(
+                    "for-{} loop body".format(itermode[0])
+                )
+                iter_ctx.replacements[itername.upper()] = item
+
+                if iteridx:
+                    iter_ctx.replacements[iteridx.upper()] = str(i)
+
+                for si, a in enumerate(f_body):
+                    yield (iter_ctx, a)
+
+        def do_else():
+            else_ctx = context.remote_derive("static for-else")
+
+            for a in f_else:
+                yield (else_ctx, a)
+
+        if itermode[0] == "group":
+            group_name = context.remote_derive(itermode[1], "a parametrized group name")
+
+            if group_name.upper() not in self.groups:
+                raise CompilerError(
+                    "No such group {} to 3 in a for loop in {}!".format(
+                        repr(group_name), context.describe()
+                    )
+                )
+
+            elif self.groups[group_name.upper()]:
+                yield from do_for(iter(self.groups[group_name.upper()]))
+
+            else:
+                yield from do_else()
+
+        elif itermode[0] == "range":
+            rang = itermode[1]
+
+            r_from = self._parse_replaceable_number(rang[0], context)
+            r_to = rang[1][0] + self._parse_replaceable_number(rang[1][1], context)
+
+            if r_to > r_from:
+                r = list(range(r_from, r_to, 1))
+                yield from do_for(r)
+
+            else:
+                yield from do_else()
+
     def _parse(self, actors):
         parsed_actors = []
 
-        context = ZDCodeParseContext(actors=[parsed_actors])
+        context = ZDCodeParseContext(actors=[parsed_actors], description="global")
 
         actors = [(context, a) for a in actors]
 
-        # unpack static for loops
-        def resolve_for(forbody):
-            itername, iteridx, itermode, f_body, f_else = forbody
-
-            def do_for(iterator):
-                break_ctx = context.derive("for")
-
-                for i, item in enumerate(iterator):
-                    iter_ctx = break_ctx.derive("for..{} loop body".format(itermode[0]))
-                    iter_ctx.replacements[itername.upper()] = item
-
-                    if iteridx:
-                        iter_ctx.replacements[iteridx.upper()] = str(i)
-
-                    for si, a in enumerate(f_body):
-                        yield iter_ctx, a
-
-            def do_else():
-                else_ctx = context.derive("for-else")
-
-                for a in f_else:
-                    yield else_ctx, a
-
-            if itermode[0] == "group":
-                group_name = context.resolve(itermode[1], "a parametrized group name")
-
-                if group_name.upper() not in self.groups:
-                    raise CompilerError(
-                        "No such group {} to 3 in a for loop in {}!".format(
-                            repr(group_name), context.describe()
-                        )
-                    )
-
-                elif self.groups[group_name.upper()]:
-                    yield from do_for(iter(self.groups[group_name.upper()]))
-
-                else:
-                    yield from do_else()
-
-            elif itermode[0] == "range":
-                rang = itermode[1]
-
-                r_from = self._parse_replaceable_number(rang[0], context)
-                r_to = rang[1][0] + self._parse_replaceable_number(rang[1][1], context)
-
-                if r_to > r_from:
-                    r = list(range(r_from, r_to, 1))
-                    yield from do_for(r)
-
-                else:
-                    yield from do_else()
-
         def unpack(vals, reslist):
             # always parse groups first,
             # to ensure they can be seen by
@@ -2426,7 +2635,7 @@ class ZDCode:
             for ctx, (class_type, a) in vals:
                 if class_type == "for":
                     unpacked = False
-                    reslist.extend(resolve_for(a))
+                    reslist.extend(self.resolve_for(a, ctx))
 
                 elif class_type != "group":
                     reslist.append((ctx, (class_type, a)))
@@ -2454,6 +2663,8 @@ class ZDCode:
             if class_type == "class template":
                 a = dict(a)
 
+                classname = self._parse_formattable_string(a["classname"], actx)
+
                 abstract_labels = set()
                 abstract_macros = {}
                 abstract_arrays = {}
@@ -2465,7 +2676,7 @@ class ZDCode:
                     if g.upper() not in self.groups:
                         raise CompilerError(
                             "Group '{}' not found while compiling template class {}!".format(
-                                g, a["classname"]
+                                g, classname
                             )
                         )
 
@@ -2495,23 +2706,24 @@ class ZDCode:
                     abstract_arrays,
                     g,
                     self,
-                    a["classname"],
+                    classname,
                     self._parse_inherit(a["inheritance"], actx),
                     a["replacement"],
                     a["class number"],
                 )
-                actx.templates[a["classname"]] = template
+                actx.templates[classname] = template
 
         pending = queue.PriorityQueue()
 
         for actx, (class_type, a) in actors:
             if class_type == "class":
                 a = dict(a)
+                classname = self._parse_formattable_string(a["classname"], actx)
 
-                with actx.desc_block("class '{}'".format(a["classname"])):
+                with actx.desc_block("class '{}'".format(classname)):
                     actor = ZDActor(
                         self,
-                        a["classname"],
+                        classname,
                         self._parse_inherit(a["inheritance"], actx),
                         a["replacement"],
                         a["class number"],
@@ -2523,12 +2735,12 @@ class ZDCode:
                         g = a["group"]
 
                         if g.upper() in self.groups:
-                            self.groups[g.upper()].append(stringify(a["classname"]))
+                            self.groups[g.upper()].append(stringify(classname))
 
                         else:
                             raise CompilerError(
                                 "Group '{}' not found while compiling class '{}'!".format(
-                                    g, a["classname"]
+                                    g, classname
                                 )
                             )
 
@@ -2545,45 +2757,40 @@ class ZDCode:
 
                     pending.put_nowait(pending_oper_gen())
 
-                    self.actors.append(actor)
+                    # self.actors.append(actor)
                     self.actor_names[actor.name.upper()] = actor
                     parsed_actors.append(actor)
 
-            elif class_type == "static template derivation":
+            if class_type == "static template derivation":
                 a = dict(a)
 
                 new_name = a["classname"]
                 gname = a["group"]
 
+                new_name = self._parse_formattable_string(new_name, actx)
+
                 with actx.desc_block(
                     "static template derivation '{}'".format(new_name)
                 ):
-                    parsed_actors.append(
-                        self._parse_template_derivation(
-                            a["source"][1],
-                            actx,
-                            pending=pending,
-                            name=new_name,
-                            do_stringify=False,
-                        )
+                    self._parse_template_derivation(
+                        a["source"][1],
+                        actx,
+                        pending=pending,
+                        name=new_name,
+                        do_stringify=False,
                     )
 
                     if gname:
+                        g = unstringify(gname)
 
-                        @pending_task(1)
-                        def pending_oper():
-                            g = unstringify(gname)
-
-                            if g.upper() not in self.groups:
-                                raise CompilerError(
-                                    "No such group {} to add the derivation {} to!".format(
-                                        repr(g), new_name
-                                    )
+                        if g.upper() not in self.groups:
+                            raise CompilerError(
+                                "No such group {} to add the derivation {} to!".format(
+                                    repr(g), new_name
                                 )
+                            )
 
-                            self.groups[g.upper()].append(new_name)
-
-                        pending.put_nowait(pending_oper)
+                        self.groups[g.upper()].append(new_name)
 
         while not pending.empty():
             pending.get_nowait().func()
@@ -2591,6 +2798,13 @@ class ZDCode:
         for a in parsed_actors:
             a.prepare_spawn_label()
 
+        self.actors.extend(parsed_actors)
+
+        self.actors.sort(key=lambda actor: actor.name)  # predominantly alphabetic sort
+        reorders = self.reorder_inherits()
+
+        print("(Reordered {} actors)".format(reorders))
+
     def __init__(self):
         self.includes = {}
         self.inventories = []
@@ -2601,9 +2815,33 @@ class ZDCode:
         self.id = make_id(35)
         self.num_anonym_macros = 0
 
+    def reorder_inherits(self) -> int:
+        new_order = []
+        positions = {}
+        reorders = 0
+
+        for actor in self.actors:
+            new_pos = len(new_order)
+
+            if actor.name in positions:
+                new_pos = positions[actor.name]
+                reorders += 1
+
+            if actor.inherit is not None:
+                if actor.inherit not in positions or positions[actor.inherit] > new_pos:
+                    positions[actor.inherit] = new_pos
+
+            new_order.insert(new_pos, actor)
+
+        self.actors = new_order
+
+        return reorders
+
     def to_decorate(self):
         res = TextNode(indent=0)
 
+        res.add_line("// :ZDCODE version='{}' id='{}' ".format(__VERSION__, self.id))
+
         if self.inventories:
             for i in self.inventories:
                 res.add_line(i.to_decorate())
